/**
 * Definitions relating to the configuration of the STAG library.
 *
 * This file is provided as part of the STAG library and released under the MIT
 * license.
 */
#ifndef STAG_TEST_STAG_H
#define STAG_TEST_STAG_H

namespace stag {
  // Provide metadata about the library version
  static const unsigned VERSION_MAJOR = 0;
  static const unsigned VERSION_MINOR = 2;
  static const unsigned VERSION_PATCH = 1;

}

#endif //STAG_TEST_STAG_H
